/*
  Copyright (C) 2006-2014 Alessandro Bugatti (alessandro.bugatti@istruzione.it)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

/*! \file
 *  \brief Come creare un  programma con OpenGl + SDL
 *  \author Alessandro Bugatti
 *  \version 0.1
 *  \date  Creazione 04/02/2014
 *  \date  Ultima modifica 17/02/2014
 *  \attention Questo programma per funzionare sotto Windows ha bisogno
 *  della presenza della libreria SDL.dll che pu essere installata
 *  nella cartella di sistema o nela cartella dell'eseguibile.
 *  Nelle distribuzioni Linux la libreria SDL di solito  installata
 *  di default.
 */
#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <sstream>
#include <GL/gl.h>
#include <GL/glu.h>
#include <SDL/SDL.h>

using namespace std;

#define SCREEN_WIDTH  640   //Larghezza dello schermo
#define SCREEN_HEIGHT 480   //Altezza dello schermo
#define SCREEN_BPP     16   //Profondit di colore

/* Superficie SDL su cui si andr a disegnare
 * Rappresenta lo schermo
 */
SDL_Surface *screen;

/* Funzione che resetta il viewport dopo un ridimensionameto della finestra */
int resizeWindow( int width, int height )
{
    /* Rapporto altezza/larghezza della finestra */
    GLfloat ratio;

    /* Controllo un'eventuale divisione per zero */
    if ( height == 0 )
	height = 1;

    ratio = ( GLfloat )width / height;

    /* Inizializzo il viewport. */
    glViewport( 0, 0, ( GLsizei )width, ( GLsizei )height );

    /* Cambio la matrice di proiezione e inizializzo il volume di visualizzazione. */
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity( );

    /* Inizializzo la prospettiva */
    gluPerspective( 45.0f, ratio, 0.1f, 100.0f );

    /* Cambio il modelview */
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity( );

    return true;
}

/* Funzione per inizializzare OpenGL */
int initGL()
{

   /* Abilita lo smooth shading , cio la capacit
     * di interpolare il colore tra vertici con colori diversi
     */
    glShadeModel( GL_SMOOTH );

    /* Inizializza lo sfondo a nero */
    glClearColor( 0.7f, 0.0f, 0.0f, 0.0f );

    /* Inizializza il Depth Buffer */
    glClearDepth( 1.0f );

    /* Abilita il DEPTH_TEST */
    glEnable( GL_DEPTH_TEST );

    /* Seleziona il tipo di Depth Test da fare */
    glDepthFunc( GL_LEQUAL );

    /* Imposta la prospettiva */
    glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );

    return true;
}

/* Codice di disegno */
int drawGLScene( )
{
    /* Serveno se si vogliono calcolare i fps (frame per second) */
    static GLint T0     = 0;
    static GLint Frames = 0;
    ostringstream out;

    /* Pulisce i buffer */
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    /* Sposta l'origine degli assi di riferimento 1,5 unit a destra
    e 6 unit dentro lo schermo
      */
    glLoadIdentity();
    glTranslatef( -1.5f, 0.0f, -6.0f );
    /* Disegna un triangolo */
    glBegin( GL_TRIANGLES );
      glVertex3f(  0.0f,  1.0f, -20.0f ); /* Punto superiore */
      glVertex3f( -1.0f, -1.0f, 0.0f ); /* In basso a sinistra */
      glVertex3f(  1.0f, -1.0f, 0.0f ); /* In basso a destra */

    glEnd( );                           /* Fine del disegno del triangolo */

    /* Muove il sistema di coordinate a destra di 3 unit */
    glTranslatef( 3.0f, 0.0f, 0.0f );
    /*Disegna un quadrato*/
    glBegin( GL_QUADS );
      glVertex3f( -1.0f,  1.0f, 0.0f ); /* In alto a sinistra */
      glVertex3f(  1.0f,  1.0f, 0.0f ); /* In alto a destra */
      glVertex3f(  1.0f, -1.0f, 0.0f ); /* In basso a destra */
      glVertex3f( -1.0f, -1.0f, 0.0f ); /* In basso a sinistra */
    glEnd( );                           /* Fine del disegno del quadrato */

    /* Disegna a schermo scambiando i buffer */
    SDL_GL_SwapBuffers( );

    /* Incrementa il numero di frame */
    Frames++;
    {
	GLint t = SDL_GetTicks();
	//Fa il calcolo ogni 5 secondi
	if (t - T0 >= 1000) {
	    GLfloat seconds = (t - T0) / 1000.0;
	    GLfloat fps = Frames / seconds;
	    out << "Frames al secondo = " << fps << " FPS\n";
        SDL_WM_SetCaption(out.str().c_str(),0);
	    T0 = t;
	    Frames = 0;
	}
    }

    return true ;
}

int main( int argc, char **argv )
{
    /* Flag da passare a SDL_SetVideoMode. Contengono informazioni sulla
       modalit video
    */
    int videoFlags;
    /* Variabile per terminare il main loop */
    int done = false;
    /* Viariabile per leggere gli eventi */
    SDL_Event event;
    /* Contiene informazioni relative al display */
    const SDL_VideoInfo *videoInfo;

    /* Inizializzo SDL */
    if ( SDL_Init( SDL_INIT_VIDEO ) < 0 )
	{
	    cerr << "Inizializzazione SDL fallita: " <<
		     SDL_GetError( ) << endl;
	    exit(1);
	}
	//Registro la funzione da chiamare in caso di terminazione
	//SDL_Quit "rimette a posto" tutto
    atexit (SDL_Quit);
    //Setto il titolo della finestra
    SDL_WM_SetCaption("Prima lezione","");
    /* Estraggo le informazioni sul video (quantit di memoria video,
       possibilit di avere alcune funzioni in hardware ecc.) */
    videoInfo = SDL_GetVideoInfo( );
    cout << "Quantit di memoria disponibile: " << videoInfo->video_mem << endl;
    if ( !videoInfo )
	{
	    cerr << "GetVideoInfo fallita: " <<
		     SDL_GetError( ) << endl;
	    exit(1);
	}

    /* Imposto i flag da passare a SDL_SetVideoMode */
    videoFlags  = SDL_OPENGL;          /* Abilita OpenGL in SDL */
    videoFlags |= SDL_GL_DOUBLEBUFFER; /* Abilita il double buffering */
    videoFlags |= SDL_HWPALETTE;       /* MEmorizza la palette in hardware */
    videoFlags |= SDL_RESIZABLE;       /* Abilita la possibilit di window resizing */

    /* Controlla se pu memorizzare lo schermo in hardware */
    if ( videoInfo->hw_available )
	   videoFlags |= SDL_HWSURFACE;
    else
	   videoFlags |= SDL_SWSURFACE;

    /* Controlla se  presente il blit hardware */
    if ( videoInfo->blit_hw )
	   videoFlags |= SDL_HWACCEL;

    /* Abilita l'OpenGL double buffering */
    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

    /* Ottiene il puntatore allo schermo */
    screen = SDL_SetVideoMode( SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_BPP,
				videoFlags );

    /* Verifica se tutto  andato bene */
    if ( !screen )
	{
	    cerr <<  "Settaggio della modalit video fallito: " << SDL_GetError();
	    exit(1);
	}

    /* Inizializzo OpenGL */
    initGL();

    /* Dimensiono la finestra iniziale */
    resizeWindow( SCREEN_WIDTH, SCREEN_HEIGHT );

    /* main loop: continuo a ridisegnare e quando arriva qualche
       evento lo processo */
    while (!done)
	{
	    //Legge la coda degli eventi e riempie la struttura event
	    while ( SDL_PollEvent( &event ) )
		{
		    switch( event.type )
			{
			case SDL_VIDEORESIZE:
			    /* Ridimensionamento della finestra */
			    screen = SDL_SetVideoMode( event.resize.w,
							event.resize.h,
							16, videoFlags );
			    if ( !screen )
				{
				    cerr << "Ridimensionamento video non riuscito: " << SDL_GetError();
				    exit(1);
				}
			    resizeWindow( event.resize.w, event.resize.h );
			    break;
			case SDL_KEYDOWN:
			    /* gestisco la pressione di un tasto */
			    switch ( event.key.keysym.sym )
            	{
            	case SDLK_ESCAPE: //premuto il tasto ESC
            	    exit(0);
            	    break;
            	default:
                    break;
            	}
			    break;
			case SDL_QUIT:
			    done = true; //Gestisco la richiesta di uscita
			    break;
			default:
			    break;
			}
		}

	    /* Disegno la scena */
	    drawGLScene( );
	}
    return 0 ;
}
