/*
  Copyright (C) 2006-2009 Alessandro Bugatti (alessandro.bugatti@istruzione.it)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

/*! \file
 *  \brief Sfere volanti: vengono disegnate una serie di sfere, tramite
 *  l'utilizzo delle quadriche definite in GLU e vengono fatte muovere
 *  verso l'utente, se escono dallo spazio di visione vengono eliminate.
 *  Con la pressione dei tasti si possono modificare alcune caratteristiche
 *  del disegno:
 *  a: accende l'illuminazione
 *  s: spegne l'illuminazione
 *  w: disegna le sfere in wireframe
 *  p: disegna le sfere come nuvole di punti
 *  f: disegna le sfere piene
 *  \author Alessandro Bugatti
 *  \version 0.2
 *  \date  Creazione 04/02/2009
 *  \date  Ultima modifica 10/01/2013
 *  \attention Questo programma per funzionare sotto Windows ha bisogno
 *  della presenza della libreria SDL.dll che pu essere installata
 *  nella cartella di sistema o nela cartella dell'eseguibile.
 *  Nelle distribuzioni Linux la libreria SDL di solito  installata
 *  di default.
 */


#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <cmath>
#include <sstream>
#include <GL/gl.h>
#include <GL/glu.h>
#include <SDL/SDL.h>
using namespace std;

#define SCREEN_WIDTH  640   //Larghezza dello schermo
#define SCREEN_HEIGHT 480   //Altezza dello schermo
#define SCREEN_BPP     16   //Profondit di colore
#define NUMERO_SFERE 50    //NUmero di sfere presenti

struct Sfera{
    /*
     * Puntatore ad un oggetto di tipo quadrica: in geometria una quadrica  ogni
     * superficie di uno spazio 3-dimensionale (o in generale N-dimensionale)
     * sui complessi rappresentata da un'equazione polinomiale
     * del secondo ordine nelle variabili spaziali (coordinate)
     */
    GLUquadricObj *quadrica;
    double raggio;
    int nLongitudine;
    int nLatitudine;
    float x, y, z;
    GLenum stile;
    float velocita;
    //componenti di colore
    unsigned char r,g,b;
};

Sfera sfere[NUMERO_SFERE];


// Luce ambiente
GLfloat LightAmbient[]  = { 0.2f, 0.2f, 0.3f, 1.0f };
// Luce diffusa
GLfloat LightDiffuse[]  = { 1.0f, 1.0f, 1.0f, 1.0f };
//Posizione della luce
GLfloat Light1Position[] = { 10.0f, 1.0f, -10.0f, 1.0f };
GLfloat Light2Position[] = { 1.0f, -10.0f, -10.0f, 1.0f };
bool toggle_light_2 = true;

GLenum actual_style = GLU_FILL;


/* Funzione per inizializzare OpenGL */
bool initGL()
{
    /*Imposto il viewport
    glViewport(x, y, larghezza, altezza)*/
    glViewport(0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
    /*Imposto il colore di sfondo, quello utilizzato per "ripitturare"
    glClearColor(rosso, verde , blu, canale_alfa)*/
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    /*Imposto il depth test*/
    glClearDepth(1.0);
    glDepthFunc(GL_LESS);
    glEnable(GL_DEPTH_TEST);
    /* Inizializzo la prospettiva */
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity( );
    GLfloat ratio = ( GLfloat )SCREEN_WIDTH / SCREEN_HEIGHT;
    gluPerspective( 45.0f, ratio, 0.1f, 2000.0f );
    /* Inizializzo la camera */
    //Abilita l'illuminazione
    glEnable(GL_LIGHTING);
    //Definisco le propriet dela luce ambiente
    glLightfv( GL_LIGHT1, GL_AMBIENT, LightAmbient );
    //Definisco le propriet dela luce diffusa
    glLightfv( GL_LIGHT1, GL_DIFFUSE, LightDiffuse );
    //Definisco la posizione della luce
    glLightfv( GL_LIGHT1, GL_POSITION, Light1Position );
    //Definisco le propriet dela luce ambiente
    glLightfv( GL_LIGHT2, GL_AMBIENT, LightAmbient );
    //Definisco le propriet dela luce diffusa
    glLightfv( GL_LIGHT2, GL_DIFFUSE, LightDiffuse );
    //Definisco la posizione della luce
    glLightfv( GL_LIGHT2, GL_POSITION, Light2Position );
    /* Abilito la prima luce */
    glEnable( GL_LIGHT1 );
    glDisable( GL_LIGHT2 );
    glMatrixMode( GL_PROJECTION );
    //glLoadIdentity();
    gluLookAt( 0, 0, 0, 0  , 0, -20, 0, 1, 0);
    return true;
}

int drawGLScene(  )
{
    /* Servono se si vogliono calcolare i fps (frame per second) */
    static GLint T0     = 0;
    static GLint Frames = 0;
    static GLint T1     = 0;
    static GLfloat angolo = 0;
    /* Pulisce i buffer */
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    GLint t = SDL_GetTicks();
	//Fa il calcolo ogni 0.05 secondi
	if (t - T0 >= 50) {
    glMatrixMode( GL_PROJECTION );
    //glLoadIdentity();
    gluLookAt( 0, 0, 0, 0  , 0, -20, sin(angolo), cos(angolo), 0);
    angolo += 0.001;
    T0 = t;
	}
    for( int i =0; i < NUMERO_SFERE; i++ )
    {
        glMatrixMode( GL_MODELVIEW );
        glLoadIdentity( );
        glTranslatef(sfere[i].x,sfere[i].y,sfere[i].z);
        gluQuadricDrawStyle(sfere[i].quadrica, sfere[i].stile);
        GLfloat ambient[]={sfere[i].r/255.0,sfere[i].g/255.0,sfere[i].b/255.0,1};
        glMaterialfv(GL_FRONT,GL_AMBIENT,ambient);
        GLfloat diffuse[]={sfere[i].r/255.0,sfere[i].g/255.0,sfere[i].b/255.0,1};
        glMaterialfv(GL_FRONT,GL_DIFFUSE,diffuse);
        GLfloat specular[]={sfere[i].r/255.0,sfere[i].g/255.0,sfere[i].b/255.0,1};
        glMaterialfv(GL_FRONT,GL_SPECULAR,specular);
        glMaterialf(GL_FRONT, GL_SHININESS,25);
        gluSphere(sfere[i].quadrica,
                   sfere[i].raggio,
                   sfere[i].nLongitudine,
                   sfere[i].nLatitudine);
    }

    /* Disegna a schermo scambiando i buffer */
    SDL_GL_SwapBuffers( );

    /* Incrementa il numero di frame */
    Frames++;
    {
	GLint t = SDL_GetTicks();
	//Fa il calcolo ogni 5 secondi
	if (t - T0 >= 5000) {
	    GLfloat seconds = (t - T0) / 1000.0;
	    GLfloat fps = Frames / seconds;
	    cout << "Frames al secondo = " << fps << " FPS\n";

	    ostringstream o;
	    o << "Sfere volanti - FPS : " << fps;
	    SDL_WM_SetCaption(o.str().c_str(),"");
	    T0 = t;
	    Frames = 0;
	}

	if (t - T1 >= 0) {

        for( int i =0; i < NUMERO_SFERE; i++ )
        {
            sfere[i].z += 0.01;
            //sfere[i].x += 0.01;
            if (sfere[i].z > 0)
                {
                    sfere[i].z = -20;
                    //sfere[i].x = 0;
                }
        }
       T1 = t;
    }
    }

    return true ;
}

void set_style(GLenum s)
{
    actual_style = s;
    for( int i =0; i < NUMERO_SFERE; i++ )
        sfere[i].stile = actual_style;
}

int main( int argc, char **argv )
{
    /* Variabile per terminare il main loop */
    int done = false;
    /* Viariabile per leggere gli eventi */
    SDL_Event event;

    /* Inizializzo SDL */
    if ( SDL_Init( SDL_INIT_VIDEO ) < 0 )
    {
        cerr << "Inizializzazione SDL fallita: " <<
        SDL_GetError( ) << endl;
        exit(1);
    }
    //Registro la funzione da chiamare in caso di terminazione
    //SDL_Quit "rimette a posto" tutto
    atexit (SDL_Quit);
    //Setto il titolo della finestra
    SDL_WM_SetCaption("Sfere volanti","");
    /*Abilito il double buffering (che comunque  abilitato di default)*/
    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
    /*Setto la modalit grafica
    SDL_SetVideoMode(larghezza, altezza, profondit_di_colore, attributi)*/
    SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 0, SDL_OPENGL | SDL_HWSURFACE);
    /* Inizializzo OpenGL */
    initGL();
    glMatrixMode(GL_MODELVIEW);
    //Inserisco 5 sfere nella lista
    srand(time(NULL));
    for (int i=0; i <NUMERO_SFERE; i++)
    {
        sfere[i].quadrica = gluNewQuadric();
        sfere[i].raggio = rand()%100/100.0;
        sfere[i].nLongitudine = rand()%20 + 4;
        sfere[i].nLatitudine = rand()%20 + 4;
        sfere[i].x = -rand()%10 + 5;
        sfere[i].y = -rand()%10 + 5;
        sfere[i].z = -rand()%50;
        sfere[i].stile = GLU_FILL;
        sfere[i].r = rand()%255;
        sfere[i].g = rand()%255;
        sfere[i].b = rand()%255;
    }
    cout << glGetString(GL_VERSION) << endl;
    /* main loop: continuo a ridisegnare e quando arriva qualche
       evento lo processo */
    while (!done)
    {
        //Legge la coda degli eventi e riempie la struttura event
        while ( SDL_PollEvent( &event ) )
        {
            switch ( event.type )
            {
            case SDL_KEYDOWN:
                /* gestisco la pressione di un tasto */
                switch ( event.key.keysym.sym )
                {
                case SDLK_ESCAPE: //premuto il tasto ESC
                    exit(0);
                    break;
                case SDLK_a: //premuto il tasto a abilita l'illuminazione
            	    glEnable(GL_LIGHTING);
            	    break;
        	    case SDLK_s: //premuto il tasto s disabilita l'illuminazione
            	    glDisable(GL_LIGHTING);
            	    break;
                case SDLK_w: //premuto il tasto w disegna in wireframe
            	    set_style(GLU_LINE);
            	    break;
        	    case SDLK_p: //premuto il tasto p disegna a punti
            	    set_style(GLU_POINT);
            	    break;
                case SDLK_f: //premuto il tasto f disegna i volumi pieni
            	    set_style(GLU_FILL);
            	    break;
                case SDLK_2: //premuto il tasto 2 abilita/disabilita la luce 2
            	    (toggle_light_2=!toggle_light_2)?glDisable(GL_LIGHT2):glEnable(GL_LIGHT2);
            	    break;
        	    default:
                    break;
                }
                break;
            case SDL_QUIT:
                done = true; //Gestisco la richiesta di uscita
                break;
            default:
                break;
            }
        }

        /* Disegno la scena */
        drawGLScene( );
    }
    return 0 ;
}
