/*
  Copyright (C) 2006-2014 Alessandro Bugatti (alessandro.bugatti@istruzione.it)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

/*! \file
 *  \brief texturing di una sfera
 *  \author Alessandro Bugatti
 *  \version 0.1
 *  \date  Creazione 04/02/2014
 *  \date  Ultima modifica 17/02/2014
 *  \attention Questo programma per funzionare sotto Windows ha bisogno
 *  della presenza della libreria SDL.dll che può essere installata
 *  nella cartella di sistema o nela cartella dell'eseguibile.
 *  Nelle distribuzioni Linux la libreria SDL di solito è installata
 *  di default.
 */
#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <sstream>
#include <GL/gl.h>
#include <GL/glu.h>
#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>

using namespace std;

#define SCREEN_WIDTH  640   //Larghezza dello schermo
#define SCREEN_HEIGHT 480   //Altezza dello schermo
#define SCREEN_BPP     16   //Profondità di colore

// Luce ambiente
GLfloat LightAmbient[]  = { 0.2f, 0.2f, 0.3f, 1.0f };
// Luce diffusa
GLfloat LightDiffuse[]  = { 1.0f, 1.0f, 1.0f, 1.0f };
//Posizione della luce
GLfloat Light1Position[] = { 0.0f, 0.0f, 0.0f, 1.0f };
GLfloat Light2Position[] = { 1.0f, -10.0f, -10.0f, 1.0f };


/* Superficie SDL su cui si andrà a disegnare
 * Rappresenta lo schermo
 */
SDL_Surface *screen;
GLuint texture;
GLUquadricObj * quadrica;

/* Funzione per caricare una bitmap come una GL texture */
int LoadGLTexture(const string& nomefile )
{
    SDL_Surface *surface;
    /* Carica la bitmap. Se non viene trovata ritorna false */
    if ( ( surface = SDL_LoadBMP( nomefile.c_str() ) ) )
        {
        /* Crea la texture e salva l'identificativo (un intero)
           nella variabile globale texture */
	    glGenTextures( 1, &texture );

	    /* Associa all'identificativo il tipo di texture */
	    glBindTexture( GL_TEXTURE_2D, texture );

	    /* Genera la texture leggendo i pixel della bitmap */
	    glTexImage2D( GL_TEXTURE_2D, 0, 3, surface->w,
			  surface->h, 0, GL_BGR,
			  GL_UNSIGNED_BYTE, surface->pixels );

	    /* Imposta il filtraggio lineare */
	    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	    /* Libera la memoria di surface */
        SDL_FreeSurface( surface );
        return true;
         }
    else
        return false;
}



/* Funzione che resetta il viewport dopo un ridimensionameto della finestra */
int resizeWindow( int width, int height )
{
    /* Rapporto altezza/larghezza della finestra */
    GLfloat ratio;
    static bool first = true;

    /* Controllo un'eventuale divisione per zero */
    if ( height == 0 )
	height = 1;

    ratio = ( GLfloat )width / height;

    /* Inizializzo il viewport. */
    glViewport( 0, 0, ( GLsizei )width, ( GLsizei )height );

    /* Cambio la matrice di proiezione e inizializzo il volume di visualizzazione. */
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity( );

    /* Inizializzo la prospettiva */
    gluPerspective( 45.0f, ratio, 0.1f, 100.0f );

    /* Cambio il modelview */
        glMatrixMode( GL_MODELVIEW );
        glLoadIdentity( );


    if (first)
    {
        //gluQuadricTexture(quadrica, true);
        //glBindTexture(GL_TEXTURE_2D, texture);
        first = false;
    }
    //glEnable( GL_LIGHT1 );
    return true;
}

/* Funzione per inizializzare OpenGL */
int initGL()
{

   /* Abilita lo smooth shading , cioè la capacità
     * di interpolare il colore tra vertici con colori diversi
     */
    glShadeModel( GL_SMOOTH );

    /* Inizializza lo sfondo a nero */
    glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );

    /* Inizializza il Depth Buffer */
    glClearDepth( 1.0f );

    /* Abilita il DEPTH_TEST */
    glEnable( GL_DEPTH_TEST );

    /* Seleziona il tipo di Depth Test da fare */
    glDepthFunc( GL_LEQUAL );

    /* Imposta la prospettiva */
    glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );

    quadrica = gluNewQuadric();
    gluQuadricTexture(quadrica, true);
    if (!LoadGLTexture("jupiter.bmp"))
        exit(1);
    /* Inizializzo la camera */
    //Abilita l'illuminazione
    glEnable(GL_LIGHTING);
    //Definisco le proprietࡤela luce ambiente
    glLightfv( GL_LIGHT1, GL_AMBIENT, LightAmbient );
    //Definisco le proprietࡤela luce diffusa
    glLightfv( GL_LIGHT1, GL_DIFFUSE, LightDiffuse );
    //Definisco la posizione della luce
    glLightfv( GL_LIGHT1, GL_POSITION, Light1Position );
    //Definisco le proprietࡤela luce ambiente
    glLightfv( GL_LIGHT2, GL_AMBIENT, LightAmbient );
    //Definisco le proprietࡤela luce diffusa
    glLightfv( GL_LIGHT2, GL_DIFFUSE, LightDiffuse );
    //Definisco la posizione della luce
    glLightfv( GL_LIGHT2, GL_POSITION, Light2Position );
    /* Abilito la prima luce */
    glEnable( GL_LIGHT1 );
    glDisable( GL_LIGHT2 );

    return true;
}

/* Codice di disegno */
int drawGLScene( )
{
    /* Serveno se si vogliono calcolare i fps (frame per second) */
    static GLint T0     = 0;
    static GLint Frames = 0;
    static GLfloat angolo = 180;
    ostringstream out;

    /* Pulisce i buffer */
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );


    /* Incrementa il numero di frame */
    Frames++;
    {
	GLint t = SDL_GetTicks();
	//Fa il calcolo ogni 5 secondi
	if (t - T0 >= 5000) {
	    GLfloat seconds = (t - T0) / 1000.0;
	    GLfloat fps = Frames / seconds;
	    out << "Frames al secondo = " << fps << " FPS\n";
        SDL_WM_SetCaption(out.str().c_str(),0);
	    T0 = t;
	    Frames = 0;
	}
    }
    glEnable(GL_TEXTURE_2D);

    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);


    //gluQuadricTexture(quadrica, true);

    //glBindTexture(GL_TEXTURE_2D, texture);
    glLoadIdentity();
    glTranslatef(0,0,-100);

    glRotatef(100,1,0,0);
    glRotatef(angolo+=0.1,0,0,1);

    gluSphere(quadrica,20, 40, 40);
    SDL_Delay(10);
    /* Disegna a schermo scambiando i buffer */
    SDL_GL_SwapBuffers( );


    return true ;
}

int main( int argc, char **argv )
{
    /* Flag da passare a SDL_SetVideoMode. Contengono informazioni sulla
       modalità video
    */
    int videoFlags;
    /* Variabile per terminare il main loop */
    int done = false;
    /* Viariabile per leggere gli eventi */
    SDL_Event event;
    /* Contiene informazioni relative al display */
    const SDL_VideoInfo *videoInfo;

    /* Inizializzo SDL */
    if ( SDL_Init( SDL_INIT_VIDEO ) < 0 )
	{
	    cerr << "Inizializzazione SDL fallita: " <<
		     SDL_GetError( ) << endl;
	    exit(1);
	}
	//Registro la funzione da chiamare in caso di terminazione
	//SDL_Quit "rimette a posto" tutto
    atexit (SDL_Quit);
    //Setto il titolo della finestra
    SDL_WM_SetCaption("Prima lezione","");
    /* Estraggo le informazioni sul video (quantità di memoria video,
       possibilità di avere alcune funzioni in hardware ecc.) */
    videoInfo = SDL_GetVideoInfo( );
    cout << "Quantità di memoria disponibile: " << videoInfo->video_mem << endl;
    if ( !videoInfo )
	{
	    cerr << "GetVideoInfo fallita: " <<
		     SDL_GetError( ) << endl;
	    exit(1);
	}

    /* Imposto i flag da passare a SDL_SetVideoMode */
    videoFlags  = SDL_OPENGL;          /* Abilita OpenGL in SDL */
    videoFlags |= SDL_GL_DOUBLEBUFFER; /* Abilita il double buffering */
    videoFlags |= SDL_HWPALETTE;       /* MEmorizza la palette in hardware */
    videoFlags |= SDL_RESIZABLE;       /* Abilita la possibilità di window resizing */

    /* Controlla se può memorizzare lo schermo in hardware */
    if ( videoInfo->hw_available )
	   videoFlags |= SDL_HWSURFACE;
    else
	   videoFlags |= SDL_SWSURFACE;

    /* Controlla se è presente il blit hardware */
    if ( videoInfo->blit_hw )
	   videoFlags |= SDL_HWACCEL;

    /* Abilita l'OpenGL double buffering */
    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

    /* Ottiene il puntatore allo schermo */
    screen = SDL_SetVideoMode( SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_BPP,
				videoFlags );

    /* Verifica se tutto è andato bene */
    if ( !screen )
	{
	    cerr <<  "Settaggio della modalità video fallito: " << SDL_GetError();
	    exit(1);
	}

    /* Inizializzo OpenGL */
    initGL();

    /* Dimensiono la finestra iniziale */
    resizeWindow( SCREEN_WIDTH, SCREEN_HEIGHT );

    /* main loop: continuo a ridisegnare e quando arriva qualche
       evento lo processo */
    while (!done)
	{
	    //Legge la coda degli eventi e riempie la struttura event
	    while ( SDL_PollEvent( &event ) )
		{
		    switch( event.type )
			{
			case SDL_VIDEORESIZE:
			    /* Ridimensionamento della finestra */
			    screen = SDL_SetVideoMode( event.resize.w,
							event.resize.h,
							16, videoFlags );
			    if ( !screen )
				{
				    cerr << "Ridimensionamento video non riuscito: " << SDL_GetError();
				    exit(1);
				}
			    resizeWindow( event.resize.w, event.resize.h );
			    break;
			case SDL_KEYDOWN:
			    /* gestisco la pressione di un tasto */
			    switch ( event.key.keysym.sym )
            	{
            	case SDLK_ESCAPE: //premuto il tasto ESC
            	    exit(0);
            	    break;
            	default:
                    break;
            	}
			    break;
			case SDL_QUIT:
			    done = true; //Gestisco la richiesta di uscita
			    break;
			default:
			    break;
			}
		}

	    /* Disegno la scena */
	    drawGLScene( );
	}
    return 0 ;
}
