/*
  Copyright (C) 2006-2014 Alessandro Bugatti (alessandro.bugatti@istruzione.it)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

/*! \file
 *  \brief Planetario versione finale
 <li>Ci sono ben 5 pianeti: Mercurio, Venere, Terra, Marte e Giove (non abbiam badato a spese)</li>
 <li>Le proporzioni e le traiettorie sono abbastanza realistiche (ovviamente a meno del Sole)</li>
 <li>Con i tasti freccia è possibile muovere la telecamera avanti e indietro su un carrello fissato all'asse Z e ruotando a destra e sinistra</li>
 <li>Con i tasti w e s è possibile inclinare il sistema solare a proprio piacimento (senza nessuno sforzo)</li>
 <li>Con i tasti a e d è possibile accelerare e decelerare l'animazione</li>
 <li>Con i tasti 1 e 2 è possibile accendere e spegnere le due luci di scena</li>
 *  \author Alessandro Bugatti
 *  \version 0.1
 *  \date  Creazione 04/02/2014
 *  \date  Ultima modifica 25/02/2014
 *  \attention Questo programma per funzionare sotto Windows ha bisogno
 *  della presenza della libreria SDL.dll che può essere installata
 *  nella cartella di sistema o nela cartella dell'eseguibile.
 *  Nelle distribuzioni Linux la libreria SDL di solito è installata
 *  di default.
 */


#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <sstream>
#include <GL/gl.h>
#include <GL/glu.h>
#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>

using namespace std;

struct OggettoCeleste
{
    char nome[30];
    char file[30];
    GLuint texture;
    float raggio;
    int paralleli;
    int meridiani;
    float rotazione_orbita;
    float raggio_rotazione;
    float rotazione_asse;
    float angolo_orbita;
    float angolo_asse;
    GLUquadricObj *quadrica;
};

#define SCREEN_WIDTH  800   //Larghezza dello schermo
#define SCREEN_HEIGHT 600   //Altezza dello schermo
#define SCREEN_BPP     16   //Profondità di colore

// Luce ambiente
GLfloat LightAmbient[]  = { 1.0f, 1.0f, 1.0f, 1.0f };
// Luce diffusa
GLfloat LightDiffuse[]  = { 1.0f, 1.0f, 1.0f, 1.0f };
//Posizione della luce
GLfloat Light1Position[] = { -10.0f, 0.0f, 0.0f, 1.0f };
GLfloat Light2Position[] = { 1.0f, -10.0f, -10.0f, 1.0f };


GLfloat camera_z = 0; //Spostamento della camera lungo l'asse z
GLfloat rotazione_y = 0; //Rotazione della camera intorno all'asse y
/* Superficie SDL su cui si andrà a disegnare
 * Rappresenta lo schermo
 */
SDL_Surface *screen;

//Vettore egli oggetti celesti
OggettoCeleste oggetti[20];
//NUmero di oggetti caricati
int oggettiPresenti;

//Percorso delle risorse
char percorsoRisorse[] = "textures/";

//Variabile per "ruotare" l'universo
float rotazione_universo = 0;

bool toggleL1 = true, toggleL2 = true;

//Identificatore della texture dell'immmagine introduttiva
GLuint textureIntro;

//Variabile che definisce un ritardo
int delay = 10;

/* Funzione per caricare una bitmap come una GL texture */
int LoadGLTexture(const string& nomefile, GLuint &texture)
{
    SDL_Surface *surface;
    /* Carica la bitmap. Se non viene trovata ritorna false */
    if ( ( surface = SDL_LoadBMP( nomefile.c_str() ) ) )
    {
        /* Associa all'identificativo il tipo di texture */
        glBindTexture( GL_TEXTURE_2D, texture );

        /* Genera la texture leggendo i pixel della bitmap */
        glTexImage2D( GL_TEXTURE_2D, 0, 3, surface->w,
                      surface->h, 0, GL_BGR,
                      GL_UNSIGNED_BYTE, surface->pixels );

        /* Imposta il filtraggio lineare */
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
        /* Libera la memoria di surface */
        SDL_FreeSurface( surface );
        return true;
    }
    else
        return false;
}

//Carica gli oggetti da file
void caricaOggetti(char nomefile[])
{
    FILE *f;
    char percorsoAssoluto[100];
    char temp[100];
    int a,b,c;
    strcpy(percorsoAssoluto,percorsoRisorse);
    strcat(percorsoAssoluto,nomefile);
    f = fopen(percorsoAssoluto,"r");
    if (f == NULL)
    {
        printf("Impossibile aprire il file dei pianeti.\n");
        exit(1);
    }
    while(fscanf(f,"%s",temp) != EOF)
    {
        strcpy(oggetti[oggettiPresenti].nome,temp);
        fscanf(f,"%s",temp);
        strcpy(oggetti[oggettiPresenti].file,temp);
        fscanf(f,"%f %d %d",&oggetti[oggettiPresenti].raggio,
               &oggetti[oggettiPresenti].meridiani,
               &oggetti[oggettiPresenti].paralleli);
        fscanf(f,"%f",&oggetti[oggettiPresenti].rotazione_orbita);
        fscanf(f,"%f",&oggetti[oggettiPresenti].raggio_rotazione);
        fscanf(f,"%f",&oggetti[oggettiPresenti].rotazione_asse);
        fscanf(f,"%f",&oggetti[oggettiPresenti].angolo_orbita);
        fscanf(f,"%f",&oggetti[oggettiPresenti].angolo_asse);
        oggettiPresenti++;
    }
    fclose(f);
}


/* Funzione che resetta il viewport dopo un ridimensionameto della finestra */
int resizeWindow( int width, int height )
{
    /* Rapporto altezza/larghezza della finestra */
    GLfloat ratio;

    /* Controllo un'eventuale divisione per zero */
    if ( height == 0 )
        height = 1;

    ratio = ( GLfloat )width / height;

    /* Inizializzo il viewport. */
    glViewport( 0, 0, ( GLsizei )width, ( GLsizei )height );

    /* Cambio la matrice di proiezione e inizializzo il volume di visualizzazione. */
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity( );

    /* Inizializzo la prospettiva */
    gluPerspective( 45.0f, ratio, 0.1f, 1000.0f );

    /* Cambio il modelview */
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity( );
    return true;
}

/* Funzione per inizializzare OpenGL */
int initGL()
{

    /* Abilita lo smooth shading , cioè la capacità
      * di interpolare il colore tra vertici con colori diversi
      */
    glShadeModel( GL_SMOOTH );

    /* Inizializza lo sfondo a nero */
    glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );

    /* Inizializza il Depth Buffer */
    glClearDepth( 1.0f );

    /* Abilita il DEPTH_TEST */
    glEnable( GL_DEPTH_TEST );

    /* Seleziona il tipo di Depth Test da fare */
    glDepthFunc( GL_LEQUAL );

    /* Imposta la prospettiva */
    glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );

    //Generazione delle quadriche e delle textures
    for (int i = 0; i < oggettiPresenti; i++)
    {
        glGenTextures(1, &oggetti[i].texture);
        oggetti[i].quadrica = gluNewQuadric();
        gluQuadricTexture(oggetti[i].quadrica, true);
        char percorsoAssoluto[100];
        strcpy(percorsoAssoluto,percorsoRisorse);
        strcat(percorsoAssoluto,oggetti[i].file);
        if (!LoadGLTexture(percorsoAssoluto,oggetti[i].texture))
            exit(1);
    }

    //Carica la texture per il titolo introduttivo
    glGenTextures(1, &textureIntro);
    if (!LoadGLTexture("textures/titolo.bmp",textureIntro))
        exit(1);

    //Abilita l'illuminazione
    glEnable(GL_LIGHTING);
    //Definisco le proprietࡤela luce ambiente
    glLightfv( GL_LIGHT1, GL_AMBIENT, LightAmbient );
    //Definisco le proprietࡤela luce diffusa
    glLightfv( GL_LIGHT1, GL_DIFFUSE, LightDiffuse );
    //Definisco la posizione della luce
    glLightfv( GL_LIGHT1, GL_POSITION, Light1Position );
    //Definisco le proprietࡤela luce ambiente
    glLightfv( GL_LIGHT2, GL_AMBIENT, LightAmbient );
    //Definisco le proprietࡤela luce diffusa
    glLightfv( GL_LIGHT2, GL_DIFFUSE, LightDiffuse );
    //Definisco la posizione della luce
    glLightfv( GL_LIGHT2, GL_POSITION, Light2Position );
    /* Abilito la prima luce */
    glEnable( GL_LIGHT1 );
    glDisable( GL_LIGHT2 );

    //Abilita il texturing
    glEnable(GL_TEXTURE_2D);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);


    return true;
}

/* Codice di disegno */
int drawGLScene( )
{
    /* Serveno se si vogliono calcolare i fps (frame per second) */
    static GLint T0     = 0;
    static GLint Frames = 0;
    GLint t;
    ostringstream out;

    /* Pulisce i buffer */
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );


    /* Incrementa il numero di frame */
    Frames++;
    t = SDL_GetTicks();
    //Fa il calcolo ogni 5 secondi
    if (t - T0 >= 5000)
    {
        GLfloat seconds = (t - T0) / 1000.0;
        GLfloat fps = Frames / seconds;
        out << "Frames al secondo = " << fps << " FPS\n";
        SDL_WM_SetCaption(out.str().c_str(),0);
        T0 = t;
        Frames = 0;
    }
    //Carico i parametri di proiezione
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    gluPerspective( 45.0f, (GLfloat)SCREEN_WIDTH/SCREEN_HEIGHT, 0.1f, 1000.0f );
    gluLookAt(0,0,camera_z,500*sin(rotazione_y),0,-500*cos(rotazione_y),0,1,0);
    /* Cambio il modelview */
    glMatrixMode( GL_MODELVIEW );

    glLoadIdentity();
    //Sposto il sistema di riferimento
    glTranslatef(0,0,-40);
    glRotatef(90+rotazione_universo,1,0,0);

    for (int i = 0; i < oggettiPresenti; i++)
    {
        glPushMatrix();
        oggetti[i].angolo_orbita += oggetti[i].rotazione_orbita;
        oggetti[i].angolo_asse += oggetti[i].rotazione_asse;
        //Applico la rotazione dell'orbita
        glRotatef(oggetti[i].angolo_orbita,0,0,1);
        //Applco la traslazione del raggio dell'orbita
        glTranslatef(oggetti[i].raggio_rotazione,0,0);
        //Applico la rotazione dell'orbita
        glRotatef(oggetti[i].angolo_asse,0,0,1);
        //Attacco la texture alla quadrica
        glBindTexture(GL_TEXTURE_2D,oggetti[i].texture);
        //Disegno il pianeta
        gluSphere(oggetti[i].quadrica,
                  oggetti[i].raggio,
                  oggetti[i].meridiani,
                  oggetti[i].paralleli);
        glPopMatrix();
    }
    SDL_Delay(delay);
    /* Disegna a schermo scambiando i buffer */
    SDL_GL_SwapBuffers( );


    return true ;
}

bool drawIntro()
{
    static int durata = 0;
    if (durata == 500)
    {
        LightAmbient[0] = 1.0 ;
        LightAmbient[1] = 1.0 ;
        LightAmbient[2] = 1.0 ;
        glLightfv( GL_LIGHT1, GL_AMBIENT, LightAmbient );
        LightDiffuse[0] = 1.0 ;
        LightDiffuse[1] = 1.0 ;
        LightDiffuse[2] = 1.0 ;
        glLightfv( GL_LIGHT1, GL_DIFFUSE, LightDiffuse );
        glEnable( GL_LIGHT2 );
    }
    if (durata >= 500)
    {
        if (durata == 500) durata++;
        return false;
    }
    LightAmbient[0] = 1.0 - 1.0*durata/450;
    LightAmbient[1] = 1.0 - 1.0*durata/450;
    LightAmbient[2] = 1.0 - 1.0*durata/450;
    glLightfv( GL_LIGHT1, GL_AMBIENT, LightAmbient );
    LightDiffuse[0] = 1.0 - 1.0*durata/450;
    LightDiffuse[1] = 1.0 - 1.0*durata/450;
    LightDiffuse[2] = 1.0 - 1.0*durata/450;
    glLightfv( GL_LIGHT1, GL_DIFFUSE, LightDiffuse );
    /* Pulisce i buffer */
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    glLightfv( GL_LIGHT1, GL_AMBIENT, LightAmbient );
    //Carico i parametri di proiezione
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    gluPerspective( 45.0f, (GLfloat)SCREEN_WIDTH/SCREEN_HEIGHT, 0.1f, 1000.0f );
    gluLookAt(0,0,camera_z,500*sin(rotazione_y),0,-500*cos(rotazione_y),0,1,0);

    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    /* Muove il sistema di coordinate a destra di 3 unità*/
    glTranslatef( 0.0f, 0.0f, -5.0f );
    glBindTexture( GL_TEXTURE_2D, textureIntro );
    /*Disegna un quadrato*/
    glBegin( GL_QUADS );

    /* In basso a sinistra */
    glTexCoord2f( 0.0f, 1.0f );
    glVertex3f( -2.0f, -1.0f, 1.0f );
    /* In basso a destra */
    glTexCoord2f( 1.0f, 1.0f );
    glVertex3f(  2.0f, -1.0f, 1.0f );
    /* In alto a destra */
    glTexCoord2f( 1.0f, 0.0f );
    glVertex3f(  2.0f,  1.0f, 1.0f );
    /* In alto a sinistra */
    glTexCoord2f( 0.0f, 0.0f );
    glVertex3f( -2.0f,  1.0f, 1.0f );


    glEnd( );
    durata++;
    SDL_Delay(10);
    /* Disegna a schermo scambiando i buffer */
    SDL_GL_SwapBuffers( );
    return true;
}

int main( int argc, char **argv )
{
    /* Flag da passare a SDL_SetVideoMode. Contengono informazioni sulla
       modalità video
    */
    int videoFlags;
    /* Variabile per terminare il main loop */
    int done = false;
    /* Viariabile per leggere gli eventi */
    SDL_Event event;
    /* Contiene informazioni relative al display */
    const SDL_VideoInfo *videoInfo;

    /* Inizializzo SDL */
    if ( SDL_Init( SDL_INIT_VIDEO ) < 0 )
    {
        cerr << "Inizializzazione SDL fallita: " <<
             SDL_GetError( ) << endl;
        exit(1);
    }
    //Registro la funzione da chiamare in caso di terminazione
    //SDL_Quit "rimette a posto" tutto
    atexit (SDL_Quit);
    //Setto il titolo della finestra
    SDL_WM_SetCaption("Planetario 1.0","");
    /* Estraggo le informazioni sul video (quantità di memoria video,
       possibilità di avere alcune funzioni in hardware ecc.) */
    videoInfo = SDL_GetVideoInfo( );
    cout << "Quantità di memoria disponibile: " << videoInfo->video_mem << endl;
    if ( !videoInfo )
    {
        cerr << "GetVideoInfo fallita: " <<
             SDL_GetError( ) << endl;
        exit(1);
    }

    /* Imposto i flag da passare a SDL_SetVideoMode */
    videoFlags  = SDL_OPENGL;          /* Abilita OpenGL in SDL */
    videoFlags |= SDL_GL_DOUBLEBUFFER; /* Abilita il double buffering */
    videoFlags |= SDL_HWPALETTE;       /* MEmorizza la palette in hardware */
    videoFlags |= SDL_RESIZABLE;       /* Abilita la possibilità di window resizing */

    /* Controlla se può memorizzare lo schermo in hardware */
    if ( videoInfo->hw_available )
        videoFlags |= SDL_HWSURFACE;
    else
        videoFlags |= SDL_SWSURFACE;
    //Abilito il fullscreen
    //videoFlags |= SDL_FULLSCREEN;
    /* Controlla se è presente il blit hardware */
    if ( videoInfo->blit_hw )
        videoFlags |= SDL_HWACCEL;

    /* Abilita l'OpenGL double buffering */
    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

    /* Ottiene il puntatore allo schermo */
    screen = SDL_SetVideoMode( SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_BPP,
                               videoFlags );

    /* Verifica se tutto è andato bene */
    if ( !screen )
    {
        cerr <<  "Settaggio della modalità video fallito: " << SDL_GetError();
        exit(1);
    }

    //Carico le informazioni sui pianeti
    caricaOggetti("pianeti.txt");

    /* Inizializzo OpenGL */
    initGL();

    /* main loop: continuo a ridisegnare e quando arriva qualche
       evento lo processo */
    while (!done)
    {
        //Legge la coda degli eventi e riempie la struttura event
        while ( SDL_PollEvent( &event ) )
        {
            switch( event.type )
            {
            case SDL_VIDEORESIZE:
                /* Ridimensionamento della finestra */
                screen = SDL_SetVideoMode( event.resize.w,
                                           event.resize.h,
                                           16, videoFlags );
                if ( !screen )
                {
                    cerr << "Ridimensionamento video non riuscito: " << SDL_GetError();
                    exit(1);
                }
                resizeWindow( event.resize.w, event.resize.h );
                break;
            case SDL_KEYDOWN:
                /* gestisco la pressione di un tasto */
                switch ( event.key.keysym.sym )
                {
                case SDLK_ESCAPE: //premuto il tasto ESC
                    exit(0);
                    break;
                case SDLK_a: //accelera
                    delay-=1;
                    if (delay < 0)
                        delay = 0;
                    break;
                case SDLK_d: //decelera
                    delay+=1;
                    break;
                case SDLK_1:
                    if (toggleL1)
                        glEnable(GL_LIGHT1);
                    else
                        glDisable(GL_LIGHT1);
                    toggleL1 = !toggleL1;
                    break;
                case SDLK_2:
                    if (toggleL2)
                        glEnable(GL_LIGHT2);
                    else
                        glDisable(GL_LIGHT2);
                    toggleL2 = !toggleL2;
                    break;
                default:
                    break;
                }
                break;
            case SDL_QUIT:
                done = true; //Gestisco la richiesta di uscita
                break;
            default:
                break;
            }
        }
        //Guardo quali tasti sono premuti e sposto la telecamera
        //Leggo lo stato della tastiera e e inserisce le informazioni
        //sui tasti premuti nel vettore keys
        Uint8 *keys = SDL_GetKeyState(NULL);
        //Camera avanti lungo l'asse z
        if ( keys[SDLK_UP] )
            camera_z-=0.1;
        //Camera indietro lungo l'asse z
        if ( keys[SDLK_DOWN] )
            camera_z+=0.1;
        //Camera ruota il target a destra
        if ( keys[SDLK_RIGHT] )
            rotazione_y+=0.01;
        //Camera ruota il target a sinistra
        if ( keys[SDLK_LEFT] )
            rotazione_y-=0.01;
        //L'universo ruota in avanti
        if ( keys[SDLK_w] )
            rotazione_universo-=0.1;
        //L'universo ruota indietro
        if ( keys[SDLK_s] )
            rotazione_universo+=0.1;
        /* Disegno la scena */
        drawGLScene( );
        while(drawIntro());
    }


    return 0 ;
}
